#include <uapi/linux/ptrace.h>
#include <linux/sched.h>
#include <linux/bpf.h>

#define WATCH_ALLOC 0
#define WATCH_FREE 1

#define MAX_FILE_NAME_LEN   32
struct type3_key_t  {
    u32 pid;
    u64 stack_id;
};

struct data_t {
    u32 type;       // 0: alloc, 1: free
    u64 size;       // 分配或释放文件缓存的大小，单位为KB
    char comm[MAX_FILE_NAME_LEN];   // 进程名或者文件名
};

BPF_HASH(cache_info, u64, struct data_t);
BPF_HASH(cache_info_stack, struct type3_key_t, struct data_t);
BPF_STACK_TRACE(cache_stack_traces, 10240);
BPF_RINGBUF_OUTPUT(events, 64);


static inline int page_cache_monitor_type3(struct pt_regs* ctx, int watch)
{
   struct type3_key_t key = {};

    key.stack_id = cache_stack_traces.get_stackid(ctx, 0);
    if(key.stack_id < 0 ) {
        bpf_trace_printk("failed to get call stack!!!!!\n");
        u64 *buf = events.ringbuf_reserve(sizeof(u64));
        if (buf) {
            events.ringbuf_submit(buf, 0);
        }
        return 0;
    }
    u64 size = PAGE_SIZE / 1024; // 单位转换为KB
    key.pid = bpf_get_current_pid_tgid();
    struct data_t *data = cache_info_stack.lookup(&key);
    if (data) {
        // 如果调用栈已经存在于哈希表中，则累加文件缓存大小
        data->size += size;
        if(key.stack_id < 0)
            bpf_trace_printk("found record %s %d %ld\n", &data->comm[0], key.pid, key.stack_id);
    } else {
        // 如果调用栈不存在于哈希表中，则创建新的数据项
        struct data_t new_data = {};
        new_data.type = watch; 
        new_data.size = size;
        bpf_get_current_comm(&new_data.comm, sizeof(new_data.comm));
        cache_info_stack.update(&key, &new_data);
        if(key.stack_id < 0)
            bpf_trace_printk("new record %s %d %ld\n", &new_data.comm[0], key.pid, key.stack_id);
    } 

    return 0;

}
int trace_add_to_page_cache_locked_type3(struct pt_regs *ctx)
{
    return page_cache_monitor_type3(ctx, WATCH_ALLOC);
}

int trace__delete_from_page_cache_type3(struct pt_regs *ctx)
{
    return page_cache_monitor_type3(ctx, WATCH_FREE);
}

static int page_cache_monitor_type2(struct inode* inode, int watch)
{
    struct dentry* dentry = NULL;    
    if( inode && inode->i_ino > 11)    {

        u64 inumber = inode->i_ino;
        u64 size = PAGE_SIZE / 1024; // 单位转换为KB

        struct data_t *data = cache_info.lookup(&inumber);
        if( data )  {
             // 如果inode已经存在于哈希表中，则累加文件缓存大小
            data->size += size;
        } else {
             // 如果inode不存在于哈希表中，则创建新的数据项
            struct data_t new_data = {};
            if(inode->i_dentry.first)   {
                char* ptr = (char*)inode->i_dentry.first;
                dentry = (struct dentry*)(ptr-offsetof(struct dentry, d_u.d_alias));
                if(dentry)  {
                    struct qstr qs = {};
                    bpf_probe_read_kernel(&qs, sizeof(qs), (void *)&dentry->d_name);
                    if (qs.len == 0)
                        return 0;
                    bpf_probe_read_kernel(&new_data.comm, sizeof(new_data.comm), (void *)qs.name);
                    new_data.type = watch; 
                    new_data.size = size;
                    cache_info.update(&inumber, &new_data);
                }            
            }
        }
    }
    return 0;
}

#if FOLIO_SUPPORT
int trace__delete_from_page_cache_folio_type2(struct pt_regs *ctx, struct folio *folio)
{
    int ret = 0;
    if(folio && folio->mapping && folio->mapping->host)
        ret = page_cache_monitor_type2(folio->mapping->host, WATCH_FREE);
    return ret;
}

int trace_add_to_page_cache_locked_folio_type2(struct pt_regs *ctx, struct address_space *mapping)
{
    int ret = 0;
    if(mapping && mapping->host)
        ret = page_cache_monitor_type2(mapping->host, WATCH_ALLOC);
    return ret;
}
#endif

int trace__delete_from_page_cache_type2(struct pt_regs *ctx, struct page *page)
{
    int ret = 0;
    if( page && page->mapping && page->mapping->host )
        ret = page_cache_monitor_type2(page->mapping->host, WATCH_FREE);
    return ret;
}

int trace_add_to_page_cache_locked_type2(struct pt_regs *ctx, struct page *page, struct address_space *mapping)
{
    int ret = 0;
    if(mapping && mapping->host)
        ret = page_cache_monitor_type2(mapping->host, WATCH_ALLOC);
    return ret;   
}


static inline int page_cache_monitor_type01(int watch)
{
    u64 pid = bpf_get_current_pid_tgid();
    u64 size = PAGE_SIZE / 1024; // 单位转换为KB
    pid = pid & 0xffff;

    struct data_t *data = cache_info.lookup(&pid);
    if (data) {
        // 如果进程已经存在于哈希表中，则累加文件缓存大小
        data->size += size;
    } else {
        // 如果进程不存在于哈希表中，则创建新的数据项
        struct data_t new_data = {};
        new_data.type = watch; 
        new_data.size = size;
        bpf_get_current_comm(&new_data.comm, sizeof(new_data.comm));
        cache_info.update(&pid, &new_data);
    }

    return 0;
}
int trace_add_to_page_cache_locked_type01(struct pt_regs *ctx)
{
    return page_cache_monitor_type01(WATCH_ALLOC);
}

inline int trace_delete_from_page_cache_type01(struct pt_regs *ctx)
{
    return page_cache_monitor_type01(WATCH_FREE);
}
