#!/usr/bin/python3
#
# opsinsight_cache Trace and display file cache allocation and release in the kernel
#
# USAGE: usage: opsinsight_cache cache [-h] [-i INTERVAL] [-t {0,1,2,3}] 
#                                     [-c {0,1,2}] [-f LOGFILE] [-r ROTATE] 
#                                     [-s LOGSIZE] [-n NUMBER] [-w {0,1}]
#
# Licensed under the Apache License, Version 2.0 (the "License")
# Copyright (C) 2024 qinyungao@kylinos.com.cn.
from bpfcc import BPF
from time import sleep, strftime
from datetime import datetime
import argparse
import configparser
from logging.handlers import RotatingFileHandler
import ctypes as ct
from subprocess import call
import os, sys, logging , threading


# display method
CONSOLE_NO_DISPLAY = 0
CONSOLE_DISPLAY = 1
CONSOLE_DISPLAY_AND_CLEANSCREN = 2

#watch type
WATCH_ALLOC = 0
WATCH_RELEASE = 1

#caculation type
TYPE_PROCESS = 0
TYPE_THREAD = 1
TYPE_FILE = 2
TYPE_CALLSTACK = 3


class cache_monitor():
    logger = None
    b = None
    parser = None
    cache_info_map = None
    stack_traces = None
    start_time = 0
    end_time = 0
    lock = threading.Lock()

    def __init__(self, args=sys.argv[1:]):
        self.cmdline = args
        self.cfg_interval = 10
        self.cfg_logfile = ""
        self.cfg_logsize = 1024
        self.cfg_rotate = 2
        self.cfg_console = 1
        self.cfg_number = 10
        self.cfg_watch = WATCH_ALLOC
        self.cfg_type = TYPE_PROCESS
 
    def get_system_boot_time(self):
        with open('/proc/stat', 'r') as file:
            for line in file:
                if line.startswith('btime'):
                    return int(line.split()[1])
    
    def get_kernel_version(self):
        with open('/proc/version', 'r') as f:
            version_info = f.read().strip().split(' ')[2]
            return version_info
    
    def compare_kernel_version(self,target_version):
        # 获取当前 Linux 内核的版本
        current_version = self.get_kernel_version()

        # 按"."分割版本号字符串
        current_version_parts = current_version.split('.')

        # 将版本号的主要和次要版本号转换为整数
        major_version = int(current_version_parts[0])
        minor_version = int(current_version_parts[1])

        # 比较版本号
        if major_version > target_version[0] or (major_version == target_version[0] and minor_version > target_version[1]):
            return True
        else:
            return False
    
    def init_cache_cfg(self):
           # read init cfg
            cfg = configparser.ConfigParser()
            cfg_name = find_ops_cfgname()
            if cfg_name == None:
                print("warning:no find cfg name;use default param\n")
            else:
                cfg.read(cfg_name)
            self.cfg_interval = int(cfg.get("cache", "interval", fallback="10"))
            self.cfg_console = int(cfg.get("cache", "console", fallback="1"))
            self.cfg_logfile = cfg.get("cache", "logfile", fallback="")
            self.cfg_rotate = int(cfg.get("cache", "rotate", fallback="2"))
            self.cfg_logsize = int(cfg.get("cache", "logsize", fallback="1024"))
            self.cfg_number = int(cfg.get("cache", "numbers", fallback="10"))
            self.cfg_watch = int(cfg.get("cache", "watch", fallback="0"))
            self.cfg_type = int(cfg.get("cache", "type", fallback="0"))

            # arguments
            examples_cache = '''examaple:
                 opsinsight cache
                 opsinsight cache -i 20 (or --interval=20)  # 指定观测周期为20s,单位为秒,默认为10s
                 opsinsight cache -f /var/log/opsinsight-cache.log (or --logfile=/var/log/opsinsight-cache.log) #指定log日志存储位置,默认不存
                 opsinsight cache -s 100 (or --logsize=100) # 指定log文件大小100MiB,单位为MiB.超过则进行轮转。默认日志大小为1024MiB
                 opsinsight cache -r 5 (or --rotete=5)      # 指定5个日志文件进行轮转,默认为2
                 opsinsight cache -c 0 (or --console=0)     # 指定不输出到屏幕(0),默认输出到屏幕(1),输出到屏幕并且清屏(2)
                 opsinsight cache -n 6 (or --numbers=6)      # 指定输出前多少项,默认为10
                 opsinsight cache -t 2 (or --type=2)        # 指定统计的方式, 0: 按进程ID进行统计, 1: 按线程ID进行统计, 2: 按文件名进行统计, 3: 按调用栈进行统计 默认为0.
                 opsinsight cache -w 1 (or --watch=1)       # 指定观测的内容, 0: 观测文件缓存占用， 1: 观测文件缓存释放, 默认为0
                 opsinsight cache -h   (or--help)             # 提供帮助信息
            '''
            self.mainparser = argparse.ArgumentParser(prog="opsinsight", description="Main command")
            subparsers = self.mainparser.add_subparsers(dest="command", title="Commands", description="Available commands", metavar="")
            parser_cache = subparsers.add_parser("cache", help="CACHE monitoring", formatter_class=argparse.RawTextHelpFormatter, epilog=examples_cache)
            parser_cache.add_argument("-i", "--interval", type=int, default=self.cfg_interval,
                                help="set observation period, Default is 10s.")
            parser_cache.add_argument("-t", "--type", type=int, choices=[0, 1, 2, 3], default=self.cfg_type,
                          help="set calculation type. Default is 0.\n"
                               "0: Count by process ID.\n"
                               "1: Count by thread ID.\n"
                               "2: Count by file name.\n"
                               "3: Count by call stack.\n")

            parser_cache.add_argument("-c", "--console", type=int, choices=[0, 1, 2], default=self.cfg_console,
                    help="Specify whether output to console. Default is 1.\n"
                    "0: Do not output to the console.\n"
                    "1: Output to the console.\n"
                    "2: Output to the console and clear the screen.\n")
                    
            parser_cache.add_argument("-f", "--logfile", type=str, default=self.cfg_logfile,
                    help="send output to file")

            parser_cache.add_argument("-r", "--rotate", type=int, default=self.cfg_rotate,
                    help="the number of files to save log. Default is 2.")
            
            parser_cache.add_argument("-s", "--logsize", type=int, default=self.cfg_logsize,
                    help="the size of the log file in MB. Default is 1024.")
            
            parser_cache.add_argument("-n", "--numbers", type=int, default=self.cfg_number,
                    help="the number of items to output, default is 10")           
            
            parser_cache.add_argument("-w", "--watch", type=int, choices=[0, 1], default=self.cfg_watch,
                    help="what to monitor. default is 0.\n"
                     "0: monitor file cache allocation\n"
                     "1: monitor file cache release")
            self.parser = parser_cache
            
    def init_cfg(self):
        try:
            self.init_cache_cfg()
            args = self.parser.parse_args(self.cmdline)
            self.cfg_interval = int(args.interval)
            if self.cfg_interval < 10:
                print("Interval cannot be less than 10, Adjusted to 10.")
                self.cfg_interval = 10
            self.cfg_console = int(args.console)
            self.cfg_logfile = str(args.logfile)
            self.cfg_rotate = int(args.rotate)
            self.cfg_logsize = int(args.logsize)
            self.cfg_number = int(args.numbers)
            self.cfg_watch = int(args.watch)
            self.cfg_type = int(args.type)
            pass
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)
            exit()

    def debug_cfg_params(self):
        print("Started CACHE Monitor, press Ctrl+C to exit.....")
        print("interval=%d"%(self.cfg_interval))
        print("console=%d"%self.cfg_console)        
        print("logfile=%s"%self.cfg_logfile)
        print("rotate=%d"%self.cfg_rotate)
        print("logsize=%d"%self.cfg_logsize)
        print("numbers=%d"%self.cfg_number)
        print("watch=%d"%self.cfg_watch)
        print("type=%d"%self.cfg_type)

    def init_log(self):
        cache_monitor.logger = logging.getLogger()
        cache_monitor.logger.setLevel(logging.INFO)
        formater = logging.Formatter('%(message)s')
        console_formater = logging.Formatter('%(message)s')
        if(self.cfg_logfile != '''\"\"'''):
            handler = RotatingFileHandler(self.cfg_logfile, maxBytes=self.cfg_logsize * 1024*1024 , backupCount=self.cfg_rotate)
            handler.setLevel(level=logging.INFO)
            handler.setFormatter(formater)
            self.handler = handler
            cache_monitor.logger.addHandler(handler)
        
        if(self.cfg_console != CONSOLE_NO_DISPLAY):
            console = logging.StreamHandler()
            console.setLevel(level=logging.INFO)
            console.setFormatter(console_formater)
            cache_monitor.logger.addHandler(console)

    def show_data_stack(self):
        cache_monitor.start_time = datetime.now()
        if(self.cfg_watch==WATCH_ALLOC):
            title="Alloc: "
        else:
            title="Release: "

        while True:
            sleep(self.cfg_interval)

            if self.cfg_console == CONSOLE_DISPLAY_AND_CLEANSCREN:
                call("clear")

            end_time = datetime.now()
            cache_monitor.logger.info("\n Start: %s" % cache_monitor.start_time.strftime("%a %b %d %H:%M:%S %Y"))
            cache_monitor.logger.info(" End  : %s" % end_time.strftime("%a %b %d %H:%M:%S %Y"))

            cache_monitor.lock.acquire()
            cache_info = cache_monitor.cache_info_map.items()
            sorted_cache_info = sorted(cache_info, key=lambda x: x[1].size, reverse=True)
            output_buffer = []  # 用于暂存要输出的内容

            for key, data in sorted_cache_info[:self.cfg_number]:
                try:
                    if key.stack_id <= cache_monitor.stack_traces.max_entries:                    
                        output_buffer.append("\n%s %s" % (title, cache_monitor.get_size_string(data.size)))
                        output_buffer.append("%s\t%d" % (data.comm.decode(), key.pid))
                        for addr in cache_monitor.stack_traces.walk(key.stack_id):
                            sym = cache_monitor.b.ksym(addr, show_offset=0).decode('utf-8', str(addr))
                            output_buffer.append("\t%s" % sym)
                except KeyboardInterrupt:
                    exit()
                except Exception as e:
                    break

            cache_monitor.lock.release()
            # 将暂存的内容一次性输出到日志
            output_str = '\n'.join(output_buffer)
            cache_monitor.logger.info(output_str)

            # 如果时间差大于1小时
            time_diff = end_time - cache_monitor.start_time
            if time_diff.total_seconds() > 3600:
                start_time = end_time
                cache_monitor.stack_traces.clear()
                cache_monitor.cache_info_map.clear()
            
    def get_size_string(size):

        size_str = "%d KB" % size

        if(size > 1024*1024):
            size /= 1024*1024
            size_str = "%.2f GB" % size

        elif(size > 1024):
            size /= 1024
            size_str = "%.2f MB" % size

        return size_str
    
    # Process events in user space
    def show_data(self):

        start_time = datetime.now()

        size_title = "Release Size"
        if(self.cfg_watch==WATCH_ALLOC):
            size_title = "Alloc Size"
        
        watch_title = "PID"
        count_title = "Process Name"
        
        if(self.cfg_type==TYPE_THREAD):
            watch_title = "TID"
            count_title = "Thread Name"
        elif(self.cfg_type==TYPE_FILE):
            watch_title = "INODE"
            count_title = "File Name"

        while True:
            sleep(self.cfg_interval)
            end_time = datetime.now()
            time_diff = end_time - start_time

            if self.cfg_console == CONSOLE_DISPLAY_AND_CLEANSCREN:
                call("clear")

            cache_monitor.logger.info("\n Start: %s" % start_time.strftime("%a %b %d %H:%M:%S %Y"))
            cache_monitor.logger.info(" End  : %s" % end_time.strftime("%a %b %d %H:%M:%S %Y"))
            cache_monitor.logger.info("+-------------------------------------------------------------------+")
            cache_monitor.logger.info("| %-32s | %-14s | %-13s |" % ( count_title, watch_title, size_title))
            cache_monitor.logger.info("|-------------------------------------------------------------------|")
            cache_info = cache_monitor.b["cache_info"].items()
            sorted_cache_info = sorted(cache_info, key=lambda x: x[1].size, reverse=True)
            for pid, data in sorted_cache_info[:self.cfg_number]:
                show_size = cache_monitor.get_size_string(data.size)
                cache_monitor.logger.info("| %-32s | %-14d | %-13s |" % (data.comm.decode(), pid.value, show_size))
            cache_monitor.logger.info("+-------------------------------------------------------------------+")

            # 如果时间差大于1小时
            if time_diff.total_seconds() > 3600:
                start_time = end_time
                cache_monitor.b["cache_info"].clear()

    def print_cache_info(self):
        if(self.cfg_type==TYPE_CALLSTACK):
            self.show_data_stack()
        else:
            self.show_data()
    
    def process_event(cpu, data, size):   
        cache_monitor.logger.info("\n Clear cache info map: %s" % cache_monitor.start_time.strftime("%a %b %d %H:%M:%S %Y"))
        cache_monitor.lock.acquire()
        cache_monitor.stack_traces.clear()
        cache_monitor.cache_info_map.clear()
        cache_monitor.start_time = datetime.now()
        cache_monitor.lock.release()
    
    def init_bpf(self):
        try:            
            with open(find_file("./", "cache_load.c"), "r") as fd:
                self.src_code = fd.read()
                fd.close()

            if(self.cfg_type == TYPE_PROCESS):
                self.src_code=self.src_code.replace("bpf_get_current_pid_tgid()", "bpf_get_current_pid_tgid()>>32")

            target_version = (5, 14)
            if self.compare_kernel_version(target_version):
                kprobe_add_func="__filemap_add_folio"
                kprobe_del_func="__filemap_remove_folio"
                add_type2_hook_func="trace_add_to_page_cache_locked_folio_type2"
                del_type2_hook_func="trace__delete_from_page_cache_folio_type2"
                self.src_code=self.src_code.replace("FOLIO_SUPPORT", "1")
            else:
                kprobe_add_func="__add_to_page_cache_locked"
                kprobe_del_func="__delete_from_page_cache"
                add_type2_hook_func="trace_add_to_page_cache_locked_type2"
                del_type2_hook_func="trace__delete_from_page_cache_type2"
                self.src_code=self.src_code.replace("FOLIO_SUPPORT", "0")

            cache_monitor.b = BPF(text=self.src_code)
            cache_monitor.stack_traces = self.b.get_table("cache_stack_traces")
            cache_monitor.cache_info_map = self.b.get_table("cache_info_stack")

            if(self.cfg_type==TYPE_THREAD or self.cfg_type==TYPE_PROCESS):
                if(self.cfg_watch==WATCH_ALLOC):
                    cache_monitor.b.attach_kprobe(event=kprobe_add_func, fn_name="trace_add_to_page_cache_locked_type01")
                else:
                    cache_monitor.b.attach_kprobe(event=kprobe_del_func, fn_name="trace_delete_from_page_cache_type01")

            if(self.cfg_type==TYPE_FILE):
                if(self.cfg_watch==WATCH_ALLOC):
                    cache_monitor.b.attach_kprobe(event=kprobe_add_func, fn_name=add_type2_hook_func)
                else:
                    cache_monitor.b.attach_kprobe(event=kprobe_del_func, fn_name=del_type2_hook_func)
            
            if(self.cfg_type == TYPE_CALLSTACK):
                if(self.cfg_watch==WATCH_ALLOC):
                    cache_monitor.b.attach_kprobe(event=kprobe_add_func, fn_name="trace_add_to_page_cache_locked_type3")
                else:
                    cache_monitor.b.attach_kprobe(event=kprobe_del_func, fn_name="trace__delete_from_page_cache_type3")
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)
            print("failed to init bpf")
            exit()    
            
    def loop(self):
        try:
            self.init_cfg()
            self.init_log()
            self.debug_cfg_params()        
            self.init_bpf()         

            # Start thread to print system load every 10 seconds
            load_thread = threading.Thread(target=self.print_cache_info)
            load_thread.daemon = True
            load_thread.start()

            cache_monitor.b["events"].open_ring_buffer(cache_monitor.process_event)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)
            exit()
        while True:
            try:
                cache_monitor.b.ring_buffer_poll()
            except KeyboardInterrupt:
                exit()

def execute(cmdline):
    engine = cache_monitor(args=cmdline)
    engine.loop()
    
if __name__ == '__main__':
    sys.path.append("../public/")
    from opstools import *
    execute(sys.argv[1:])
    # engine = cache_monitor()
    # engine.loop()
