#include <uapi/linux/ptrace.h>
#include <linux/sched.h>
#include <linux/bpf.h>

struct cpu_load_info_summary_t  {
    u64 timestamp;			//时间戳
    s64 tasks;				//系统处于就绪态和不可中断阻塞太线程之和
};

struct cpu_load_info_detail_t  {
    u64 state; 					//线程状态
    u64 cpu;
    char comm[TASK_COMM_LEN];	//线程名
    u64 stack_id; 				//调用栈标识
};

BPF_HASH(cpu_cur_load_info, u32, struct cpu_load_info_detail_t);
BPF_STACK_TRACE(stack_traces, 10240);
BPF_ARRAY(cpu_load, struct cpu_load_info_summary_t, 1);
BPF_ARRAY(cpu_load_max, struct cpu_load_info_summary_t, 1);
BPF_RINGBUF_OUTPUT(events, 1024);

int enqueue_task_fair_hook(struct pt_regs *ctx, struct rq *rq, struct task_struct *p, int flags)
{

    u32 tid = p->pid;
    
    // 忽略idle线程
    if (tid == 0)
        return 0;

    // 尝试从cpu_cur_load_info中查找给定的TID
    struct cpu_load_info_detail_t *load_info = cpu_cur_load_info.lookup(&tid);
    if (load_info) {
        // 如果存在,更新state信息
        #ifdef CPU_LOAD_DEBUG
        bpf_trace_printk("Thread ID %u found in cpu_cur_load_info. Old State 0x%llx, New State 0x%llx.", tid, load_info->state, p->state);
        #endif
        load_info->cpu = bpf_get_smp_processor_id();
        load_info->state = p->STATE_FIELD;
        load_info->stack_id = stack_traces.get_stackid(ctx, 0);
        return 0;
    } else {
        // 如果不存在,创建新的记录并插入到cpu_cur_load_info中
        struct cpu_load_info_detail_t new_load_info = {};
        new_load_info.state = p->STATE_FIELD;
        new_load_info.cpu = bpf_get_smp_processor_id();
        bpf_probe_read_kernel_str(new_load_info.comm, sizeof(new_load_info.comm), p->comm);

        ###STACK###new_load_info.stack_id = stack_traces.get_stackid(ctx, 0);
        cpu_cur_load_info.update(&tid, &new_load_info);
        
        // 打印相关信息
        #ifdef CPU_LOAD_DEBUG
        bpf_trace_printk("New record Thread ID %u, Thread state: 0x%llx, Thread name: %s",
                         tid, new_load_info.state, new_load_info.comm);
        #endif
    }

    // 更新当前系统负载
    int key_0=0;
    struct cpu_load_info_summary_t *cur_load = cpu_load.lookup(&key_0);
    if( cur_load )  {
        #ifdef CPU_LOAD_DEBUG
        bpf_trace_printk("increase cur_load=%p %d", cur_load, cur_load->tasks);
        #endif
        __sync_fetch_and_add(&cur_load->tasks, 1);
    }

    // 更新最高系统负载
    struct cpu_load_info_summary_t *max_load = cpu_load_max.lookup(&key_0);
    if (max_load && cur_load && cur_load->tasks > max_load->tasks) {
        max_load->tasks = cur_load->tasks;
        max_load->timestamp = bpf_ktime_get_ns();

        // Reserve space in ring buffer and submit data
        void *buf = events.ringbuf_reserve(sizeof(struct cpu_load_info_summary_t));
        if (buf) {
            memcpy(buf, max_load, sizeof(struct cpu_load_info_summary_t));
            events.ringbuf_submit(buf, 0);
        }
        // bpf_trace_printk("max_load tasks: %d", max_load->tasks);
    }

    return 0;
}


int dequeue_task_fair_hook(struct pt_regs *ctx, struct rq *rq, struct task_struct *p, int flags)
{
    u32 tid = p->pid;;

    // 忽略idle线程
    if (tid == 0)
        return 0;

    // 尝试从cpu_cur_load_info中查找给定的TID
    struct cpu_load_info_detail_t *load_info = cpu_cur_load_info.lookup(&tid);
    if (load_info) {
        // 如果线程状态为不可中断阻塞态,则更新state信息
        if (p->STATE_FIELD == TASK_UNINTERRUPTIBLE) {
            load_info->state = p->STATE_FIELD;
            #ifdef CPU_LOAD_DEBUG
            bpf_trace_printk("Update record Thread ID %u, Thread state: 0x%llx\n", tid, p->STATE_FIELD);
            #endif
        } else {
            // 如果线程状态不是不可中断阻塞态，则删除该记录并减小系统负载
            #ifdef CPU_LOAD_DEBUG
            bpf_trace_printk("Delete record Thread ID %u  Thread state: 0x%llx\n", tid, p->STATE_FIELD);
            #endif
            cpu_cur_load_info.delete(&tid);

            // 减小当前系统负载
            int key = 0;
            struct cpu_load_info_summary_t *cur_load = cpu_load.lookup(&key);
            if (cur_load) {
                #ifdef CPU_LOAD_DEBUG
                bpf_trace_printk("decrease cur_load=%p %d", cur_load, cur_load->tasks);
                #endif
                __sync_fetch_and_sub(&cur_load->tasks,1);
            }
        }
    }

    return 0;

}