#!/usr/bin/python3
#
# opsinsight_cpu Trace and capture the highest system load detail in the kernel
#
# usage: opsinsight_cpu cpu [-h] [-i INTERVAL] [-d {0,1,2}] [-c {0,1,2}] [-f LOGFILE] 
#                       [-r ROTATE] [-s LOGSIZE] [-l LIMIT]
#
# Licensed under the Apache License, Version 2.0 (the "License")
# Copyright (C) 2024 qinyungao@kylinos.com.cn.
from bpfcc import BPF
from time import sleep, strftime
from datetime import datetime
from datetime import datetime
import argparse
import configparser
import logging
import threading
from logging.handlers import RotatingFileHandler
import ctypes as ct
from subprocess import call
import os
import sys 

# display method
CONSOLE_NO_DISPLAY = 0
CONSOLE_DISPLAY = 1
CONSOLE_DISPLAY_AND_CLEANSCREN = 2


# Define the structures
class Data(ct.Structure):
    _fields_ = [("timestamp", ct.c_ulonglong),
                ("tasks", ct.c_long)]
    
class cpu_monitor():
    logger = None
    b = None
    parser = None
    cpu_cur_load_info = None
    stack_traces = None
    cur_system_load = []
    max_system_load = []

    max_system_load_timestap=0
    max_load=0
    lock = threading.Lock()

    def __init__(self, args=sys.argv[1:]):
        self.cmdline = args

        self.cfg_interval = 10
        self.cfg_detail = 0
        self.cfg_logfile = ""
        self.cfg_logsize = 1024
        self.cfg_rotate = 2
        self.cfg_console = 1
        self.cfg_limit = 0        

    def get_system_boot_time(self):
        with open('/proc/stat', 'r') as file:
            for line in file:
                if line.startswith('btime'):
                    return int(line.split()[1])
    
    def get_kernel_version(self):
        with open('/proc/version', 'r') as f:
            version_info = f.read().strip().split(' ')[2]
            return version_info
    
    def compare_kernel_version(self,target_version):
        # 获取当前 Linux 内核的版本
        current_version = self.get_kernel_version()

        # 按"."分割版本号字符串
        current_version_parts = current_version.split('.')

        # 将版本号的主要和次要版本号转换为整数
        major_version = int(current_version_parts[0])
        minor_version = int(current_version_parts[1])

        # 比较版本号
        if major_version > target_version[0] or (major_version == target_version[0] and minor_version > target_version[1]):
            return True
        else:
            return False
        
    def init_cpu_cfg(self):
           # read init cfg 
            cfg = configparser.ConfigParser()
            cfg_name = find_ops_cfgname()
            if cfg_name == None:
                print("warning:no find cfg name;use default param\n")
            else:
                cfg.read(cfg_name)
            self.cfg_interval = int(cfg.get("cpu", "interval", fallback="10"))
            self.cfg_detail = int(cfg.get("cpu", "detail", fallback="0"))
            self.cfg_console = int(cfg.get("cpu", "console", fallback="1"))
            self.cfg_logfile = cfg.get("cpu", "logfile", fallback="")
            self.cfg_rotate = int(cfg.get("cpu", "rotate", fallback="2"))
            self.cfg_logsize = int(cfg.get("cpu", "logsize", fallback="1024"))
            self.cfg_limit = int(cfg.get("cpu", "limit", fallback="0"))

            # arguments
            examples_cpu = '''examaple:
                 opsinsight cpu
                 opsinsight cpu -i 20 (or --interval=20)  # 指定观测周期为20s,单位为秒,默认为10s
                 opsinsight cpu -d 1 (or --detail=1)      # 指定观测的详细程度为1,默认为0
                 opsinsight cpu -f /var/log/cpu_monitor-cpu.log (or --logfile=/var/log/cpu_monitor-cpu.log) #指定log日志存储位置,默认不存
                 opsinsight cpu -s 100 (or --logsize=100) # 指定log文件大小100MiB,单位为MiB.超过则进行轮转。默认日志大小为1024MiB
                 opsinsight cpu -r 5                      # 指定5个日志文件进行轮转,默认为2
                 opsinsight cpu -c 0 (or --console=0)     # 指定不输出到屏幕(0),默认输出到屏幕(1),输出到屏幕并且清屏(2)
                 opsinsight cpu -h   (or--help)           # 提供帮助信息
            '''
            self.mainparser = argparse.ArgumentParser(prog="opsinsight", description="Main command")
            subparsers = self.mainparser.add_subparsers(dest="command", title="Commands", description="Available commands", metavar="")
            parser_cpu = subparsers.add_parser("cpu", help="CPU monitoring", formatter_class=argparse.RawTextHelpFormatter, epilog=examples_cpu)
            parser_cpu.add_argument("-i", "--interval", type=int, default=self.cfg_interval,
                                help="set observation period, Default is 10s.")
            parser_cpu.add_argument("-d", "--detail", type=int, choices=[0, 1, 2], default=self.cfg_detail,
                    help="set detail level.\n"
                    "0: Display the number of threads in R and D states at the time of highest system load.\n"
                    "1: Display the number of threads in R and D states at the time of highest system load along with thread names.\n"
                    "2: Display the number of threads in R and D states at the time of highest system load along with thread names and call stack.\n")
            parser_cpu.add_argument("-c", "--console", type=int, choices=[0, 1, 2], default=self.cfg_console,
                    help="Specify whether output to console. Default is 1.\n"
                    "0: Do not output to the console.\n"
                    "1: Output to the console.\n"
                    "2: Output to the console and clear the screen.\n")
                    
            parser_cpu.add_argument("-f", "--logfile", type=str, default=self.cfg_logfile,
                    help="send output to file")

            parser_cpu.add_argument("-r", "--rotate", type=int, default=self.cfg_rotate,
                    help="the number of files to save log. Default is 2.")
            
            parser_cpu.add_argument("-s", "--logsize", type=int, default=self.cfg_logsize,
                    help="the size of the log file in MB. Default is 1024.")

            parser_cpu.add_argument("-l", "--limit", type=int, default=self.cfg_limit,
                    help="Specify the threshold for displaying load information. Default is 0.\n")
            self.parser = parser_cpu
            
    def init_bpf(self):
        try:
            filename = find_file("./", "cpu_load.c")
            with open(filename, "r") as fd:
                self.src_code = fd.read()
                fd.close()

            target_version = (5, 14)
            if self.compare_kernel_version(target_version):
                self.src_code = self.src_code.replace("STATE_FIELD","__state")
            else:
                self.src_code = self.src_code.replace("STATE_FIELD","state")
                
            # Start polling for events
            if(self.cfg_detail==2):
                self.src_code = self.src_code.replace("###STACK###","")
            else:
                self.src_code = self.src_code.replace("###STACK###","//")
            
            cpu_monitor.b = BPF(text=self.src_code)
            cpu_monitor.b.attach_kprobe(event="enqueue_task_fair", fn_name="enqueue_task_fair_hook")
            cpu_monitor.b.attach_kprobe(event="dequeue_task_fair", fn_name="dequeue_task_fair_hook")
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print("failed to init bpf")
            exit()

    def init_cfg(self):
        try:
            self.init_cpu_cfg()
            args = self.parser.parse_args(self.cmdline)
            self.cfg_interval = int(args.interval)
            if self.cfg_interval < 10:
                print("Interval cannot be less than 10, Adjusted to 10.")
                self.cfg_interval = 10
            self.cfg_detail = int(args.detail)
            self.cfg_console = int(args.console)
            self.cfg_logfile = str(args.logfile)
            self.cfg_rotate = int(args.rotate)
            self.cfg_logsize = int(args.logsize)
            self.cfg_limit = int(args.limit)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)
            exit()

    def debug_cfg_params(self):
        print("Started CPU Monitor, press Ctrl+C to exit.....")
        print("interval=%d"%(self.cfg_interval))
        print("detail=%d"%self.cfg_detail)
        print("console=%d"%self.cfg_console)

        print("logfile=%s"%self.cfg_logfile)
        print("rotate=%d"%self.cfg_rotate)
        print("logsize=%d"%self.cfg_logsize)
        print("limit=%d"%self.cfg_limit)

    def init_log(self):
        cpu_monitor.logger = logging.getLogger()
        cpu_monitor.logger.setLevel(logging.INFO)
        #formater = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        formater = logging.Formatter('%(message)s')
        console_formater = logging.Formatter('%(message)s')
        if(self.cfg_logfile != '''\"\"'''):
            handler = RotatingFileHandler(self.cfg_logfile, maxBytes=self.cfg_logsize * 1024*1024 , backupCount=self.cfg_rotate)
            handler.setLevel(level=logging.INFO)
            handler.setFormatter(formater)
            self.handler = handler
            cpu_monitor.logger.addHandler(handler)
        
        if(self.cfg_console != CONSOLE_NO_DISPLAY):
            console = logging.StreamHandler()
            console.setLevel(level=logging.INFO)
            console.setFormatter(console_formater)
            cpu_monitor.logger.addHandler(console)

    # Function to print system load
    def print_system_load(self):
        boot_time = self.get_system_boot_time()
        start_time = datetime.now()

        while True:
            sleep(self.cfg_interval)
            timestamp = datetime.fromtimestamp((cpu_monitor.max_system_load_timestap / 1e9) + boot_time).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]

            # 定义一个空的字符串列表
            output_r = []
            output_d = []
            pid_list = []
            # 循环处理数据并将输出内容添加到列表中
            cpu_monitor.lock.acquire()
            show_system_load = cpu_monitor.max_system_load.copy()
            cpu_monitor.lock.release()
            cpu_monitor.max_load=0
            cpu_monitor.b["cpu_load_max"].clear()

            for pid, info in show_system_load:
                #过滤掉opsinsight本身
                thread_name=info.comm.decode()
                if("opsinsight" in thread_name):
                    continue

                if(self.cfg_detail==2):
                    #过滤无效的调用栈信息
                   if(info.stack_id > cpu_monitor.stack_traces.max_entries):
                        print("invalid stack_id 0x%lx %ld" % (info.stack_id, info.stack_id))
                        continue

                #过滤重复的PID
                if pid.value in pid_list:
                    continue

                pid_list.append(pid.value)                

                if info.state != 0x02:                    
                    output_r.append((info.cpu, thread_name, pid.value, info.stack_id))
                else:
                    output_d.append((info.cpu, thread_name, pid.value, info.stack_id))            

            if self.cfg_console == CONSOLE_DISPLAY_AND_CLEANSCREN:
                call("clear")

            total_load = len(output_r) + len(output_d)
            if(total_load < self.cfg_limit):
                continue

            # 一次性输出所有内容
            cpu_monitor.logger.info("\n%-40s%-30s%-30s" % ("Timestamp", "Running Tasks", "Uninterruptable Tasks"))
            cpu_monitor.logger.info("%-40s%-30d%-30d" % (timestamp, len(output_r), len(output_d)))

            if(self.cfg_detail>0):
                for value in output_r:
                    cpu_monitor.logger.info("R-%d-%s-%d" % value[:3] )

                    #打印调用栈
                    if(self.cfg_detail==2):
                        cpu_monitor.logger.info("stack(%d):" % value[3])
                        for addr in cpu_monitor.stack_traces.walk(value[3]):
                            sym = cpu_monitor.b.ksym(addr, show_offset=0).decode('utf-8', 'replace')
                            cpu_monitor.logger.info("\t%s" % sym)
                        cpu_monitor.logger.info("")


                for value in output_d:
                    cpu_monitor.logger.info("D-%d-%s-%d" % value[:3] )

                    #打印调用栈
                    if(self.cfg_detail==2):
                        cpu_monitor.logger.info("stack(%d):" % value[3])
                        for addr in cpu_monitor.stack_traces.walk(value[3]):
                            sym = cpu_monitor.b.ksym(addr, show_offset=0).decode('utf-8', 'replace')
                            cpu_monitor.logger.info("\t%s" % sym)
                        cpu_monitor.logger.info("")
            
            end_time = datetime.now()
            time_diff = end_time - start_time
            if time_diff.total_seconds() > 3600:
                start_time = end_time
                cpu_monitor.logger.info("\n!!!!!!clear monitor info!!!!!!!\n")
                cpu_monitor.cur_system_load.clear()
                if(self.cfg_detail==2):
                    cpu_monitor.stack_traces.clear()
                cpu_monitor.b["cpu_load_max"].clear()
                cpu_monitor.b["cpu_load"].clear()


    # Process events in user space
    def print_event(cpu, data, size):
        event = ct.cast(data, ct.POINTER(Data)).contents
        cpu_monitor.max_system_load_timestap = event.timestamp;

    #   print("Tasks: %d" % (event.tasks))
    #   start_time = time.perf_counter_ns()  # 记录开始时间（纳秒级别）
        cpu_monitor.cur_load=len(cpu_monitor.cpu_cur_load_info)
        if(cpu_monitor.cur_load > cpu_monitor.max_load):
            cpu_monitor.cur_system_load = cpu_monitor.cpu_cur_load_info.items()
        
        if(len(cpu_monitor.cur_system_load) > cpu_monitor.max_load):
            if(cpu_monitor.lock.acquire(blocking=False)):
                cpu_monitor.max_system_load=cpu_monitor.cur_system_load.copy()
                cpu_monitor.max_load=len(cpu_monitor.max_system_load)
                cpu_monitor.lock.release()

    def loop(self):
        try:
            self.init_cfg()
            self.init_log()
            self.debug_cfg_params()        
            self.init_bpf()

            cpu_monitor.cpu_cur_load_info = self.b.get_table("cpu_cur_load_info")
            if(self.cfg_detail==2):
                cpu_monitor.stack_traces = self.b.get_table("stack_traces")
            cpu_monitor.b["events"].open_ring_buffer(cpu_monitor.print_event)            

            # Start thread to print system load every 10 seconds
            load_thread = threading.Thread(target=self.print_system_load)
            load_thread.daemon = True
            load_thread.start()
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)
            exit()

        while True:
            try:
                cpu_monitor.b.ring_buffer_poll()
            except KeyboardInterrupt:
                exit()

def execute(cmdline):
    engine = cpu_monitor(args=cmdline)
    engine.loop()    

if __name__ == '__main__':
    sys.path.append("../public/")
    from opstools import *
    execute(sys.argv[1:])
    # engine = cpu_monitor()
    # engine.loop()
