#!/usr/bin/python3

from bpfcc import BPF
import time
import sys
import os
from datetime import datetime
from subprocess import call
import argparse
import configparser
import logging
from logging.handlers import RotatingFileHandler

# display method
CONSOLE_NO_DISPLAY = 0
CONSOLE_DISPLAY = 1
CONSOLE_DISPLAY_AND_CLEANSCREN = 2

# logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger()
logger.setLevel(logging.INFO)

class mem_allloc_free_class():
    global logger

    def __init__(self, args=sys.argv[1:]):                         
        try:
            self.cfg_dict = {  #type  default min max
                "cfg_interval": ["int", 1,  1, 10000],
                "cfg_watch": ["int", 0, 0, 1],
                "cfg_type": ["int", 0, 0, 3],
                "cfg_numbers": ["int", 10, 1, 1000],
                "cfg_accu": ["int", 1, 0, 1 ],
                "cfg_logfile": ["str", "/var/log/tmp.log"],
                "cfg_logrotate": ["int", 2, 1, 10],
                "cfg_logsize": ["int", 1024, 1, 1024],
                "cfg_console": ["int", 1, 0, 2]
            }
            self.cmdline = args
            self.cfg_interval = self.cfg_dict["cfg_interval"][1]
            self.cfg_watch = self.cfg_dict["cfg_watch"][1]
            self.cfg_type = self.cfg_dict["cfg_type"][1]
            self.cfg_numbers = self.cfg_dict["cfg_numbers"][1]
            self.cfg_accu = self.cfg_dict["cfg_accu"][1]
            self.cfg_logfile = self.cfg_dict["cfg_logfile"][1]
            self.cfg_logrotate = self.cfg_dict["cfg_logrotate"][1]
            self.cfg_logsize = self.cfg_dict["cfg_logsize"][1]
            self.cfg_console = self.cfg_dict["cfg_console"][1]
            
            self.__init_cfg()
            self.init_log()
            self.check_cfg()

            filename = find_file("./", "memory.c")
            with open(filename, "r") as fd:
                self.src_code = fd.read()
                fd.close()
                if self.cfg_watch == 0:
                    self.src_code = self.src_code.replace("MACRO_GET_ID", "key.id = bpf_get_current_pid_tgid() >> 32;")
                else:
                    self.src_code = self.src_code.replace("MACRO_GET_ID", "key.id = bpf_get_current_pid_tgid();")
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            # logger.error(e)
            print(e)

    def __init_cfg(self):
        try:
            # read init cfg
            cfg = configparser.ConfigParser()
            cfg_name = find_ops_cfgname()
            if cfg_name == None:
                print("warning:no find cfg name;use default param\n")
            else:
                cfg.read(cfg_name)
            self.cfg_interval = int(cfg.get("mem", "interval", fallback=str(self.cfg_interval)))
            self.cfg_watch = int(cfg.get("mem", "watch", fallback=str(self.cfg_watch)))
            self.cfg_type = int(cfg.get("mem", "type", fallback=str(self.cfg_type)))
            self.cfg_numbers = int(cfg.get("mem", "numbers", fallback=str(self.cfg_numbers)))
            self.cfg_accu = int(cfg.get("mem", "accu", fallback=str(self.cfg_accu)))
            self.cfg_logfile = cfg.get("mem", "logfile", fallback=self.cfg_logfile)
            self.cfg_logrotate = int(cfg.get("mem", "rotate", fallback=str(self.cfg_logrotate)))
            self.cfg_logsize = int(cfg.get("mem", "logsize", fallback=str(self.cfg_logsize)))
            self.cfg_console = int(cfg.get("mem", "console", fallback=str(self.cfg_console)))

            # arguments
            examples = '''examaple:
                 ./opsinsight mem
                 ./opsinsight mem -i 2 (or --interval=2)    # 指定观测周期为2s，单位为秒，默认为1s
                 ./opsinsight mem -c 0 (or --console=0)     # 指定不输出到屏幕(0),默认输出到屏幕(1),输出到屏幕并且清屏(2)
                 ./opsinsight mem -w 1 (or --watch=1)       # 指定按线程显示（1），默认按进程显示（0）
                 ./opsinsight mem -t 1 (or --type=1)        # 指定排序方式为变化量的绝对值排序（1），默认按分配总大小排序（0）
                 ./opsinsight mem -n 10 (or --numbers=10)   # 指定显示行数，默认是10行
                 ./opsinsight mem -a 1 (or --accu=1)        # 指定统计模式，0：累积模式 1：非累积模式(默认)
                 ./opsinsight mem -f /var/log/mem.log (or --logfile=/var/log/opsinsight-mem.log) #指定log日志存储位置，默认不存
                 ./opsinsight mem -r 2 (or --rotate=2)      #
                 ./opsinsight mem -s 100 (or --logsize=100) # 指定log文件大小100MiB，单位为MiB.超过则清除日志，重新记录。默认日志大小为1024MiB

                 ./opsinsight mem -h   (or--help)           # 提供帮助信息
            '''
            main_parser = argparse.ArgumentParser(prog="opsinsight", description="Main command")            
            subparsers = main_parser.add_subparsers(dest="command", title="Commands", description="Available commands", metavar="")
            parser = subparsers.add_parser(
                "memory alloc/free process", help="memory alloc/free monitor",                
                formatter_class= argparse.RawTextHelpFormatter, epilog=examples)
            parser.add_argument("-i", "--interval", type=int, default=self.cfg_interval,
                                help="Set observation period;unit:second;default=1s")
            parser.add_argument("-w", "--watch", type=int, choices=[0,1], default=self.cfg_watch,
                                help="show pid or tid:default pid(0).\n"
                                "0:pid.\n"
                                "1:tid.\n")
            parser.add_argument("-t", "--type", type=int, choices=[0,1,2,3], default=self.cfg_type,
                                help="display sort method: default 1\n"
                                "0: alloc_size + free_size.\n"
                                "1: abs(alloc_size - free_size).\n"
                                "2: alloc_size.\n"
                                "3: free_size.")
            parser.add_argument("-n", "--numbers", type=int, default=self.cfg_numbers,
                                help="set display lines, default=%d"%self.cfg_numbers)
            parser.add_argument("-a", "--accu", type=int, default=self.cfg_accu,
                                help="set acculate display mode:"
                                "0 acculate display"
                                "1: no acculate display（default）")
            parser.add_argument("-f", "--logfile", default=self.cfg_logfile,
                                help="set logfile position")
            parser.add_argument("-r", "--rotate", type=int, default=self.cfg_logrotate,
                                help="set log rotate times: default=2")
            parser.add_argument("-s", "--logsize", type=int, default=self.cfg_logsize,
                                help="set log size;unit MiB;default 10MiB")
            parser.add_argument("-c", "--console", type=int, choices=[0, 1, 2], default=self.cfg_console,
                    help="Specify whether output to console. Default is 1.\n"
                    "0: Do not output to the console.\n"
                    "1: Output to the console.\n"
                    "2: Output to the console and clear the screen.\n")

            # parser.add_argument("-h", "---help", help="help info")
            args = parser.parse_args(self.cmdline)
            self.cfg_interval = int(args.interval)
            self.cfg_watch = int(args.watch)
            self.cfg_type = int(args.type)
            self.cfg_numbers = int(args.numbers)
            self.cfg_accu = int(args.accu)
            self.cfg_logfile = str(args.logfile)
            self.cfg_logrotate = int(args.rotate)
            self.cfg_logsize = int(args.logsize)
            self.cfg_console = int(args.console)
        except KeyboardInterrupt:
           exit()
        except Exception as e:
            logger.warning(e)

    def check_cfg(self):
        try:
            for k, val in self.cfg_dict.items():
                data = getattr(self, k)
                if val[0] == "int":
                    if data < val[2] or data > val[3]:
                        print("Warn:%s(%d) must be in %d~%d;will set default=%d"%(k, data, val[2], val[3], val[1]))
                        setattr(self, k, val[1])
                elif val[0] == "str":
                    pass
                elif val[0] == "bool":
                    pass
                else:
                    pass
        except KeyboardInterrupt:
           exit()                
        except Exception as e:
            print(e)

    def debug_cfg_params(self):
        print("interval=%d"%(self.cfg_interval))
        print("watch:%d"%self.cfg_watch)
        print("type=%d"%self.cfg_type)
        print("numbers=%d"%self.cfg_numbers)
        print("accu=%d"%self.cfg_accu)
        print("logfile=%s"%self.cfg_logfile)
        print("log_rotate=%d"%(self.cfg_logrotate))
        print("logsize=%d"%self.cfg_logsize)
        print("console=%d"%self.cfg_console)

    def init_log(self):
        formater = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_formater = logging.Formatter('%(message)s')
        # handler = logging.FileHandler(self.cfg_logfile)
        if self.cfg_logfile != '''\"\"''':
            handler = RotatingFileHandler(self.cfg_logfile, maxBytes=self.cfg_logsize * 1024*1024 , backupCount= self.cfg_logrotate)
            handler.setLevel(level=logging.INFO)
            handler.setFormatter(formater)
            logger.addHandler(handler)
            self.handler = handler

        console = logging.StreamHandler()
        console.setLevel(level=logging.INFO)
        console.setFormatter(console_formater)
        logger.addHandler(console)
        self.console = console

    def loop(self):
        try:
            self.debug_cfg_params()
            b = BPF(text=self.src_code)
            start_time = 0
            while 1:
                # init
                sg = {}
                item = {}
                # get start timestamp
                if start_time == 0 or self.cfg_accu == 1:
                    start_time = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S")
                sg["start_time"] = start_time

                # sleep
                time.sleep(self.cfg_interval)
                # get data from ebpf
                alloc_size = 0
                free_size  = 0
                for k, v in b["mem_info"].items():
                    item[k.id] = [k.id, v.alloc_size + v.free_size, abs(v.alloc_size - v.free_size),
                                  v.alloc_size, v.free_size,
                                  v.comm.decode("utf-8")]
                    alloc_size += v.alloc_size
                    free_size += v.free_size
                sg["start_time"] = sg["start_time"] + ("%16s%8s"%("alloc_size:", format_number(alloc_size)))
                # get end timestamp
                sg["end_time"] = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S") + ("%16s%8s"%("free_size:", format_number(free_size)))
                # clean statistic info
                if self.cfg_accu == 1:
                    info = b.get_table("mem_info")
                    info.clear()
                # sort data from ebpf
                sg["data"] = dict(sorted(item.items(), key=lambda item: item[1][self.cfg_type + 1], reverse=True))
                # show data
                self.show_data(dict_data=sg)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            # logger.error(e)
            print("6")
            print(e)

    def show_data(self, dict_data):
        try:
            display_lines = int(self.cfg_numbers)
            id_name = "PID"
            if self.cfg_watch:
                id_name = "TID"

            if self.cfg_console == CONSOLE_NO_DISPLAY:
                logger.removeHandler(self.console)
            elif self.cfg_console == CONSOLE_DISPLAY_AND_CLEANSCREN:
                call("clear")
            else:
                logger.info("")

            logger.info("%16s:%16s"%("Start-Time", dict_data["start_time"]))
            logger.info("%16s:%16s"%("End-Time", dict_data["end_time"]))
            if self.cfg_type == 0:
                logger.info("%8s%16s%16s%16s%16s%16s"%(id_name, "ALLOC_SIZE", "FREE_SIZE", "<TOTAL_SIZE>", "ABS_SIZE", "COMMANDS"))
            elif self.cfg_type == 1:
                logger.info("%8s%16s%16s%16s%16s%16s"%(id_name, "ALLOC_SIZE", "FREE_SIZE", "TOTAL_SIZE", "<ABS_SIZE>", "COMMANDS"))
            elif self.cfg_type == 2:
                logger.info("%8s%16s%16s%16s%16s%16s"%(id_name, "<ALLOC_SIZE>", "FREE_SIZE", "TOTAL_SIZE", "ABS_SIZE", "COMMANDS"))
            elif self.cfg_type == 3:
                logger.info("%8s%16s%16s%16s%16s%16s"%(id_name, "ALLOC_SIZE", "<FREE_SIZE>", "TOTAL_SIZE", "ABS_SIZE", "COMMANDS"))
            for k, v in dict_data["data"].items():
                logger.info("%8s%16s%16s%16s%16s%16s"%(v[0], format_number(v[3]), format_number(v[4]), format_number(v[1]),
                                                       format_number(v[2]), v[5]))
                display_lines -= 1
                if display_lines <= 0:
                    break
            if self.cfg_console == CONSOLE_NO_DISPLAY:
                logger.addHandler(self.console)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            # logger.error(e)
            print(e)

def execute(cmdline):
    cls = mem_allloc_free_class(args=cmdline)
    cls.loop()


if __name__ == '__main__':
    sys.path.append("../public/")
    from opstools import *
    execute(sys.argv[1:])
    # a = mem_allloc_free_class()
    # # a.debug_cfg_params()
    # a.loop()
