#!/usr/bin/python3

from bpfcc import BPF
import time
import os
import sys
from datetime import datetime
from subprocess import call
import argparse
import configparser
import logging
from logging.handlers import RotatingFileHandler

# display method
CONSOLE_NO_DISPLAY = 0
CONSOLE_DISPLAY = 1
CONSOLE_DISPLAY_AND_CLEANSCREN = 2

# logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger()
logger.setLevel(logging.INFO)

sched_rt_fifo = 1
sched_rt_rr = 2
sched_idle = 5
username_dict = {}

def init_username_dic():
    try:
        with open("/etc/passwd") as fd:
            while True:
                line = fd.readline()
                if line:
                    items = line.split(":")
                    if len(items) < 4:
                        continue
                    key = "%08x" % (int(items[2]))
                    username_dict[key] = items[0]
                else:
                    break
        fd.close()
    except KeyboardInterrupt:
        exit()
    except Exception as e:
        print(e)

def get_username(key):
    for k, v in username_dict.items():
        if k == key:
            return username_dict[k]
    return "??"

def format_number(i_d):
    try:
        # byte
        if i_d < int(1024):
            o_d = "%dB" % (i_d)
        # KB
        elif i_d < 0x100000:
            o_d = "%.2fK" % (i_d/1024.0)
        # MB
        elif i_d < 0x3f600000:
            o_d = "%.2fM" % (i_d/1024/1024.0)
        # GB
        else:
            o_d = '%.2fG' % (i_d/1024/1024/1024.0)
        return o_d
    except KeyboardInterrupt:
        exit()
    except Exception as e:
        print(e)
        return "%dB--" % (i_d)


def format_number_io(i_d):
    try:
        # byte
        if i_d < int(1024):
            o_d = "%d" % (i_d)
        # KB
        elif i_d < 0x100000:
            o_d = "%.2fK" % (i_d/1024.0)
        # MB
        elif i_d < 0x3f600000:
            o_d = "%.2fM" % (i_d/1024/1024.0)
        # GB
        else:
            o_d = '%.2fG' % (i_d/1024/1024/1024.0)
        return o_d
    except Exception:
        exit()
    except Exception as e:
        print(e)
        return "%dB--" % (i_d)


class io_class():
    global logger

    def __init__(self, args=sys.argv[1:]):
        try:
            self.cfg_dict = {
                # type  default min max
            "cfg_interval": ["int", 1,  1, 10000],
                "cfg_watch": ["int", 0, 0, 1],
                "cfg_type": ["int", 0, 0, 2],
                "cfg_numbers": ["int", 10, 1, 1000],
                "cfg_accu": ["int", 1, 0, 1],
                "cfg_logfile": ["str", "/var/log/tmp.log"],
                "cfg_logrotate": ["int", 2, 1, 10],
                "cfg_logsize": ["int", 10, 1, 1024],
                "cfg_console": ["int", 1, 0, 2],
                "cfg_mode": ["int", 0, 0, 1]
            }
            self.cmdline = args
            self.cfg_interval = self.cfg_dict["cfg_interval"][1]
            self.cfg_watch = self.cfg_dict["cfg_watch"][1]
            self.cfg_type = self.cfg_dict["cfg_type"][1]
            self.cfg_numbers = self.cfg_dict["cfg_numbers"][1]
            self.cfg_accu = self.cfg_dict["cfg_accu"][1]
            self.cfg_logfile = self.cfg_dict["cfg_logfile"][1]
            self.cfg_logrotate = self.cfg_dict["cfg_logrotate"][1]
            self.cfg_logsize = self.cfg_dict["cfg_logsize"][1]
            self.cfg_console = self.cfg_dict["cfg_console"][1]
            self.cfg_mode = self.cfg_dict["cfg_mode"][1]            
            self.__init_cfg()
            self.init_log()
            self.check_cfg()

            filename = find_file("./", "ops_io.c")
            with open(filename, "r") as fd:
                self.src_code = fd.read()
                fd.close()
                # key.id = bpf_get_current_pid_tgid();
                # key.id = bpf_get_current_pid_tgid() >> 32;
                # MACRO_GET_ID
                # src_code = src_code.replace("MACRO_GET_ID", "key.id = bpf_get_current_pid_tgid() >> 32;")
                # who.id = bpf_get_current_pid_tgid() >> 32;
                if self.cfg_watch == 0:
                    self.src_code = self.src_code.replace("MACRO_GET_ID", "key.id = bpf_get_current_pid_tgid()>>32;")
                    self.src_code = self.src_code.replace("MACRO_GET_DISK_ID", "who->id = bpf_get_current_pid_tgid()>>32;")
                else:
                    self.src_code = self.src_code.replace("MACRO_GET_ID", "key.id = bpf_get_current_pid_tgid();")
                    self.src_code = self.src_code.replace("MACRO_GET_DISK_ID", "who->id = bpf_get_current_pid_tgid();")
                # print(src_code)
                if self.is_kylin_352_version():
                    self.src_code = self.src_code.replace("__RQ_DISK__", "rq_disk")
                else:
                    self.src_code = self.src_code.replace("__RQ_DISK__", "q->disk")
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            # logger.error(e)
            print(e)

    def __init_cfg(self):
        try:
            # read init cfg 
            cfg = configparser.ConfigParser()
            cfg_name = find_ops_cfgname()
            if cfg_name == None:
                print("warning:no find cfg name;use default param\n")
            else:
                cfg.read(cfg_name)
            self.cfg_interval = int(cfg.get("io", "interval", fallback=str(self.cfg_interval)))
            self.cfg_watch = int(cfg.get("io", "watch", fallback=str(self.cfg_watch)))
            self.cfg_type = int(cfg.get("io", "type", fallback=str(self.cfg_type)))
            self.cfg_numbers = int(cfg.get("io", "numbers", fallback=str(self.cfg_numbers)))
            self.cfg_accu = int(cfg.get("io", "accu", fallback=str(self.cfg_accu)))
            self.cfg_logfile = cfg.get("io", "logfile", fallback=self.cfg_logfile)
            self.cfg_logrotate = int(cfg.get("io", "rotate", fallback=str(self.cfg_logrotate)))
            self.cfg_logsize = int(cfg.get("io", "logsize", fallback=str(self.cfg_logsize)))
            self.cfg_console = int(cfg.get("io", "console", fallback=str(self.cfg_console)))
            self.cfg_mode = int(cfg.get("io", "mode", fallback=str(self.cfg_mode)))

            # arguments
            examples = '''examaple:
                 ./opsinsight io
                 ./opsinsight io -i 2 (or --interval=2)    # 指定观测周期，单位为秒，默认为1s
                 ./opsinsight io -w 1 (or --watch=1)       # 指定按线程显示（1），默认按进程显示（0）
                 ./opsinsight io -t 1 (or --type=1)        # 指定排序方式， 0:按io写排序（默认）1：按io读排序， 2：按io排序
                 ./opsinsight io -n 10 (or --numbers=10)   # 指定显示行数，默认是10行
                 ./opsinsight mem -a 1 (or --accu=1)       # 指定统计模式，0：累积模式 1：非累积模式（默认）
                 ./opsinsight io -f /var/log/mem.log (or --logfile=/var/log/mem.log) #指定log日志存储位置，默认不存
                 ./opsinsight io -r 2 (or --rotate=2)      #
                 ./opsinsight io -s 100 (or --logsize=100) # 指定log文件大小100MiB，单位为MiB.默认日志大小为1024MiB
                 ./opsinsight io -c 0 (or --console=0)     # 指定不输出到屏幕（0），默认输出到屏幕（1）
                 ./opsinsight io -m 0 (or --mode=0)        # 指定采样模式；0:基于vfs采样（默认） 1：基于磁盘io请求采样
                 ./opsinsight io -h   (or--help)           # 提供帮助信息
            '''
            parser = argparse.ArgumentParser(
                description="memory alloc/free process",
                formatter_class=argparse.RawDescriptionHelpFormatter,
                epilog=examples)
            parser.add_argument("-i", "--interval", type=int, default=self.cfg_interval,
                                help="Set observation period;unit:second;default=1s")
            parser.add_argument("-w", "--watch", type=int, choices=[0,1], default=self.cfg_watch,
                                help="show pid or tid:default pid(0).\n"
                                "0:pid.\n"
                                "1:tid.\n")
            parser.add_argument("-t", "--type", type=int, choices=[0,1,2], default=self.cfg_type,
                                help="display sort method: default 1\n"
                                "0: write size.\n"
                                "1: read size.\n"
                                "2: io size.\n")
            parser.add_argument("-n", "--numbers", type=int, default=self.cfg_numbers,
                                help="set display lines:10")
            parser.add_argument("-a", "--accu", type=int, default=self.cfg_accu,
                                help="set acculate display mode:"
                                "0 acculate display"
                                "1: no acculate display（default）")
            parser.add_argument("-f", "--logfile", default=self.cfg_logfile,
                                help="set logfile position")
            parser.add_argument("-r", "--rotate", type=int, default=self.cfg_logrotate,
                                help="set log rotate times: default=2")
            parser.add_argument("-s", "--logsize", type=int, default=self.cfg_logsize,
                                help="set log size;unit MiB;default 1024MiB")
            parser.add_argument("-m", "--mode", type=int, choices=[0,1], default=self.cfg_mode,
                                help="set collect mode:default file(0).\n"
                                "0:file io.\n"
                                "1:disk io.\n")
            parser.add_argument("-c", "--console", type=int, choices=[0, 1, 2], default=self.cfg_console,
                    help="Specify whether output to console. Default is 1.\n"
                    "0: Do not output to the console.\n"
                    "1: Output to the console.\n"
                    "2: Output to the console and clear the screen.\n")

            # parser.add_argument("-h", "---help", help="help info")
            args = parser.parse_args(self.cmdline)
            self.cfg_interval = int(args.interval)
            self.cfg_watch = int(args.watch)
            self.cfg_type = int(args.type)
            self.cfg_numbers = int(args.numbers)
            self.cfg_accu = int(args.accu)
            self.cfg_logfile = str(args.logfile)
            self.cfg_logrotate = int(args.rotate)
            self.cfg_logsize = int(args.logsize)
            self.cfg_console = int(args.console)
            self.cfg_mode = int(args.mode)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            logger.warning(e)

    def check_cfg(self):
        try:
            for k, val in self.cfg_dict.items():
                data = getattr(self, k)
                if val[0] == "int":
                    if data < val[2] or data > val[3]:
                        print("Warn:%s(%d) must be in %d~%d;will set default=%d"%(k, data, val[2], val[3], val[1]))
                        setattr(self, k, val[1])
                elif val[0] == "str":
                    pass
                elif val[0] == "bool":
                    pass
                else:
                    pass
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)

    def debug_cfg_params(self):
        print("interval=%d"%(self.cfg_interval))
        print("watch:%d"%self.cfg_watch)
        print("type=%d"%self.cfg_type)
        print("numbers=%d"%self.cfg_numbers)
        print("accu=%d"%self.cfg_accu)
        print("logfile=%s"%self.cfg_logfile)
        print("log_rotate=%d"%(self.cfg_logrotate))
        print("logsize=%d"%self.cfg_logsize)
        print("console=%d"%self.cfg_console)
        print("mode=%d"%self.cfg_mode)

    def init_log(self):
        formater = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        console_formater = logging.Formatter('%(message)s')
        # handler = logging.FileHandler(self.cfg_logfile)
        if self.cfg_logfile != '''\"\"''':
            handler = RotatingFileHandler(self.cfg_logfile, maxBytes=self.cfg_logsize * 1024*1024 , backupCount= self.cfg_logrotate)
            handler.setLevel(level=logging.INFO)
            handler.setFormatter(formater)
            logger.addHandler(handler)
            self.handler = handler
            
        console = logging.StreamHandler()
        console.setLevel(level=logging.INFO)
        console.setFormatter(console_formater)
        logger.addHandler(console)        
        self.console = console
        

    def is_kylin_352_version(self):
        if BPF.get_kprobe_functions(b'blk_account_io_done'):
            return True
        else:
            return False

    def loop(self):
        try:
            self.debug_cfg_params()
            
            init_username_dic()
            b = BPF(text=self.src_code)
            if self.cfg_mode:
                # kylin-3.5-2
                if self.is_kylin_352_version():
                    b.attach_kprobe(event_re="blk_account_io_start",
                                    fn_name="blk_account_io_start_hook")
                    b.attach_kprobe(event_re="blk_account_io_done",
                                    fn_name="blk_account_io_done_hook")
                else:
                    # kylin-3.6-1
                    b.attach_tracepoint(tp="block:block_io_start", fn_name="trace_pid_start_tp")
                    b.attach_kprobe(event="blk_mq_start_request", fn_name="trace_req_start")
                    b.attach_tracepoint(tp="block:block_io_done", fn_name="trace_req_completion_tp")
            else:
                b.attach_kprobe(event_re="vfs_write", fn_name="vfs_write_hook")
                b.attach_kprobe(event_re="vfs_read", fn_name="vfs_read_hook")
            start_time = 0
            while 1:
                # init
                sg = {}
                item = {}
                # get start timestamp
                if start_time == 0 or self.cfg_accu == 1:
                    start_time = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S")
                sg["start_time"] = start_time

                # sleep
                time.sleep(self.cfg_interval)
                # get data from ebpf
                for k, v in b["io_info"].items():
                    key_username="%08x"%(v.uid)
                    username = username_dict[key_username]
                    if v.policy == sched_rt_fifo or v.policy == sched_rt_rr:
                        sched_prio = "rt/"
                    elif v.policy == sched_idle:
                        sched_prio = 'idle'
                    else:
                        sched_prio = 'be/'
                    sched_prio += "%d"%((v.prio + 20)/5)
                    item[k.id] = [k.id, username, sched_prio, v.write_bytes, v.read_bytes, v.io, v.comm.decode("utf-8")]

                # get end timestamp
                sg["end_time"] = datetime.strftime(datetime.now(), "%Y-%m-%d %H:%M:%S")
                # clean statistic info
                if self.cfg_accu == 1:
                    info = b.get_table("io_info")
                    info.clear()

                # sort data from ebpf
                sg["data"] = dict(sorted(item.items(), key=lambda item: item[1][self.cfg_type + 3], reverse=True))
                # show data
                self.show_data(dict_data=sg)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            # logger.error(e)
            print(e)

    def show_data(self, dict_data):
        try:
            display_lines = int(self.cfg_numbers)
            id_name = "PID"
            if self.cfg_watch:
                id_name = "TID"

            if self.cfg_console == CONSOLE_NO_DISPLAY:
                logger.removeHandler(self.console)
            elif self.cfg_console == CONSOLE_DISPLAY_AND_CLEANSCREN:
                call("clear")
            else:
                logger.info("")

            logger.info("%16s:%16s"%("Start-Time", dict_data["start_time"]))
            logger.info("%16s:%16s"%("End-Time", dict_data["end_time"]))
            if self.cfg_type == 0:
                logger.info("%8s%8s%8s%8s%8s%8s%16s"%(id_name, "USR", "PRIO", "<WRITE>", "READ", "IO", "COMMANDS"))
            elif self.cfg_type == 1:
                logger.info("%8s%8s%8s%8s%8s%8s%16s"%(id_name, "USR", "PRIO", "WRITE", "<READ>", "IO", "COMMANDS"))
            elif self.cfg_type == 2:
                logger.info("%8s%8s%8s%8s%8s%8s%16s"%(id_name, "USR", "PRIO", "WRITE", "READ", "<IO>", "COMMANDS"))
            for k, v in dict_data["data"].items():
                logger.info("%8s%8s%8s%8s%8s%8s%16s"%(v[0], v[1], v[2], format_number(v[3]), format_number(v[4]),
                                                           format_number_io(v[5]), v[6]))
                display_lines -= 1
                if display_lines <= 0:
                    break
            if self.cfg_console == CONSOLE_NO_DISPLAY:
                logger.addHandler(self.console)
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            print(e)

def execute(cmdline):
    cls = io_class(args=cmdline)
    cls.loop()


if __name__ == '__main__':
    sys.path.append("../public/")
    from opstools import *
    execute(sys.argv[1:])
    # a = io_class()
    # # a.debug_cfg_params()
    # a.loop()
